@@ -26,7 +26,7 @@
  * <refsect2>
  * <title>Example launch line</title>
  * |[
- * gst-launch -v v4l2src ! jpegddec ! ffmpegcolorspace ! xvimagesink
+ * gst-launch -v v4l2src ! jpegdec ! ffmpegcolorspace ! xvimagesink
  * ]| The above pipeline reads a motion JPEG stream from a v4l2 camera
  * and renders it to the screen.
  * </refsect2>
@@ -52,11 +52,13 @@
         (((struct GstJpegDecSourceMgr*)((cinfo_ptr)->src))->dec)
 
 #define JPEG_DEFAULT_IDCT_METHOD	JDCT_FASTEST
+#define JPEG_DEFAULT_MAX_ERRORS 	0
 
 enum
 {
   PROP_0,
-  PROP_IDCT_METHOD
+  PROP_IDCT_METHOD,
+  PROP_MAX_ERRORS
 };
 
 /* *INDENT-OFF* */
@@ -189,7 +191,23 @@ gst_jpeg_dec_class_init (GstJpegDecClass * klass)
   g_object_class_install_property (gobject_class, PROP_IDCT_METHOD,
       g_param_spec_enum ("idct-method", "IDCT Method",
           "The IDCT algorithm to use", GST_TYPE_IDCT_METHOD,
-          JPEG_DEFAULT_IDCT_METHOD, G_PARAM_READWRITE));
+          JPEG_DEFAULT_IDCT_METHOD,
+          G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS));
+
+  /**
+   * GstJpegDec:max-errors
+   *
+   * Error out after receiving N consecutive decoding errors
+   * (-1 = never error out, 0 = automatic, 1 = fail on first error, etc.)
+   *
+   * Since: 0.10.27
+   **/
+  g_object_class_install_property (gobject_class, PROP_MAX_ERRORS,
+      g_param_spec_int ("max-errors", "Maximum Consecutive Decoding Errors",
+          "Error out after receiving N consecutive decoding errors "
+          "(-1 = never fail, 0 = automatic, 1 = fail on first error)",
+          -1, G_MAXINT, JPEG_DEFAULT_MAX_ERRORS,
+          G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS));
 
   gstelement_class->change_state =
       GST_DEBUG_FUNCPTR (gst_jpeg_dec_change_state);
@@ -198,6 +216,81 @@ gst_jpeg_dec_class_init (GstJpegDecClass * klass)
   GST_DEBUG_CATEGORY_GET (GST_CAT_PERFORMANCE, "GST_PERFORMANCE");
 }
 
+static void
+gst_jpeg_dec_clear_error (GstJpegDec * dec)
+{
+  g_free (dec->error_msg);
+  dec->error_msg = NULL;
+  dec->error_line = 0;
+  dec->error_func = NULL;
+}
+
+static void
+gst_jpeg_dec_set_error_va (GstJpegDec * dec, const gchar * func, gint line,
+    const gchar * debug_msg_format, va_list args)
+{
+#ifndef GST_DISABLE_GST_DEBUG
+  gst_debug_log_valist (GST_CAT_DEFAULT, GST_LEVEL_WARNING, __FILE__, func,
+      line, (GObject *) dec, debug_msg_format, args);
+#endif
+
+  g_free (dec->error_msg);
+  if (debug_msg_format)
+    dec->error_msg = g_strdup_vprintf (debug_msg_format, args);
+  else
+    dec->error_msg = NULL;
+
+  dec->error_line = line;
+  dec->error_func = func;
+}
+
+static void
+gst_jpeg_dec_set_error (GstJpegDec * dec, const gchar * func, gint line,
+    const gchar * debug_msg_format, ...)
+{
+  va_list va;
+
+  va_start (va, debug_msg_format);
+  gst_jpeg_dec_set_error_va (dec, func, line, debug_msg_format, va);
+  va_end (va);
+}
+
+static GstFlowReturn
+gst_jpeg_dec_post_error_or_warning (GstJpegDec * dec)
+{
+  GstFlowReturn ret;
+  int max_errors;
+
+  ++dec->error_count;
+  max_errors = g_atomic_int_get (&dec->max_errors);
+
+  if (max_errors < 0) {
+    ret = GST_FLOW_OK;
+  } else if (max_errors == 0) {
+    /* FIXME: do something more clever in "automatic mode" */
+    if (dec->packetized) {
+      ret = (dec->error_count < 3) ? GST_FLOW_OK : GST_FLOW_ERROR;
+    } else {
+      ret = GST_FLOW_ERROR;
+    }
+  } else {
+    ret = (dec->error_count < max_errors) ? GST_FLOW_OK : GST_FLOW_ERROR;
+  }
+
+  GST_INFO_OBJECT (dec, "decoding error %d/%d (%s)", dec->error_count,
+      max_errors, (ret == GST_FLOW_OK) ? "ignoring error" : "erroring out");
+
+  gst_element_message_full (GST_ELEMENT (dec),
+      (ret == GST_FLOW_OK) ? GST_MESSAGE_WARNING : GST_MESSAGE_ERROR,
+      GST_STREAM_ERROR, GST_STREAM_ERROR_DECODE,
+      g_strdup (_("Failed to decode JPEG image")), dec->error_msg,
+      __FILE__, dec->error_func, dec->error_line);
+
+  dec->error_msg = NULL;
+  gst_jpeg_dec_clear_error (dec);
+  return ret;
+}
+
 static boolean
 gst_jpeg_dec_fill_input_buffer (j_decompress_ptr cinfo)
 {
@@ -211,6 +304,11 @@ gst_jpeg_dec_fill_input_buffer (j_decompress_ptr cinfo)
   GST_DEBUG_OBJECT (dec, "fill_input_buffer: fast av=%u, remaining=%u", av,
       dec->rem_img_len);
 
+  if (av == 0) {
+    GST_DEBUG_OBJECT (dec, "Out of data");
+    return FALSE;
+  }
+
   if (dec->rem_img_len < av)
     av = dec->rem_img_len;
   dec->rem_img_len -= av;
@@ -234,11 +332,29 @@ gst_jpeg_dec_init_source (j_decompress_ptr cinfo)
 static void
 gst_jpeg_dec_skip_input_data (j_decompress_ptr cinfo, glong num_bytes)
 {
-  GST_DEBUG_OBJECT (CINFO_GET_JPEGDEC (cinfo), "skip %ld bytes", num_bytes);
+  GstJpegDec *dec = CINFO_GET_JPEGDEC (cinfo);
+
+  GST_DEBUG_OBJECT (dec, "skip %ld bytes", num_bytes);
 
   if (num_bytes > 0 && cinfo->src->bytes_in_buffer >= num_bytes) {
     cinfo->src->next_input_byte += (size_t) num_bytes;
     cinfo->src->bytes_in_buffer -= (size_t) num_bytes;
+  } else if (num_bytes > 0) {
+    gint available;
+
+    num_bytes -= cinfo->src->bytes_in_buffer;
+    cinfo->src->next_input_byte += (size_t) cinfo->src->bytes_in_buffer;
+    cinfo->src->bytes_in_buffer = 0;
+
+    available = gst_adapter_available (dec->adapter);
+    if (available < num_bytes || available < dec->rem_img_len) {
+      GST_WARNING_OBJECT (dec, "Less bytes to skip than available in the "
+          "adapter or the remaining image length %ld < %d or %u",
+          num_bytes, available, dec->rem_img_len);
+    }
+    num_bytes = MIN (MIN (num_bytes, available), dec->rem_img_len);
+    gst_adapter_flush (dec->adapter, num_bytes);
+    dec->rem_img_len -= num_bytes;
   }
 }
 
@@ -322,6 +438,7 @@ gst_jpeg_dec_init (GstJpegDec * dec)
 
   /* init properties */
   dec->idct_method = JPEG_DEFAULT_IDCT_METHOD;
+  dec->max_errors = JPEG_DEFAULT_MAX_ERRORS;
 
   dec->adapter = gst_adapter_new ();
 }
@@ -344,8 +461,11 @@ gst_jpeg_dec_ensure_header (GstJpegDec * dec)
     return FALSE;
   }
 
+  if (offset > 0) {
+    GST_LOG_OBJECT (dec, "Skipping %u bytes.", offset);
+    gst_adapter_flush (dec->adapter, offset);
+  }
   GST_DEBUG_OBJECT (dec, "Found JPEG header");
-  gst_adapter_flush (dec->adapter, offset);
 
   return TRUE;
 }
@@ -358,9 +478,10 @@ gst_jpeg_dec_parse_tag_has_entropy_segment (guint8 tag)
   return FALSE;
 }
 
-/* returns image length in bytes if parsed
- * successfully, otherwise 0 */
-static guint
+/* returns image length in bytes if parsed successfully,
+ * otherwise 0 if more data needed,
+ * if < 0 the absolute value needs to be flushed */
+static gint
 gst_jpeg_dec_parse_image_data (GstJpegDec * dec)
 {
   guint size;
@@ -421,8 +542,15 @@ gst_jpeg_dec_parse_image_data (GstJpegDec * dec)
       dec->parse_resync = FALSE;
       dec->parse_offset = 0;
       return (offset + 4);
+    } else if (value == 0xd8) {
+      /* Skip this frame if we found another SOI marker */
+      GST_DEBUG ("0x%08x: SOI marker before EOI, skipping", offset + 2);
+      dec->parse_resync = FALSE;
+      dec->parse_offset = 0;
+      return -(offset + 2);
     }
 
+
     if (value >= 0xd0 && value <= 0xd7)
       frame_len = 0;
     else {
@@ -474,6 +602,7 @@ gst_jpeg_dec_parse_image_data (GstJpegDec * dec)
       if (noffset < 0) {
         /* ignore and continue resyncing until we hit the end
          * of our data or find a sync point that looks okay */
+        offset++;
         continue;
       }
       GST_DEBUG ("found sync at 0x%x", offset + 2);
@@ -918,10 +1047,9 @@ gst_jpeg_dec_decode_direct (GstJpegDec * dec, guchar * base[3],
 
 format_not_supported:
   {
-    GST_ELEMENT_ERROR (dec, STREAM, DECODE,
-        (_("Failed to decode JPEG image")),
-        ("Unsupported subsampling schema: v_samp factors: %u %u %u",
-            v_samp[0], v_samp[1], v_samp[2]));
+    gst_jpeg_dec_set_error (dec, GST_FUNCTION, __LINE__,
+        "Unsupported subsampling schema: v_samp factors: %u %u %u",
+        v_samp[0], v_samp[1], v_samp[2]);
     return GST_FLOW_ERROR;
   }
 }
@@ -1124,7 +1252,8 @@ gst_jpeg_dec_chain (GstPad * pad, GstBuffer * buf)
 #endif
   guchar *outdata;
   guchar *base[3], *last[3];
-  guint img_len, outsize;
+  gint img_len;
+  guint outsize;
   gint width, height;
   gint r_h, r_v;
   guint code, hdr_ok;
@@ -1162,6 +1291,7 @@ gst_jpeg_dec_chain (GstPad * pad, GstBuffer * buf)
     goto need_more_data;
   }
 
+again:
   if (!gst_jpeg_dec_ensure_header (dec))
     goto need_more_data;
 
@@ -1176,23 +1306,28 @@ gst_jpeg_dec_chain (GstPad * pad, GstBuffer * buf)
      * is not aligned to buffer boundaries */
     img_len = gst_jpeg_dec_parse_image_data (dec);
 
-    if (img_len == 0)
+    if (img_len == 0) {
       goto need_more_data;
+    } else if (img_len < 0) {
+      gst_adapter_flush (dec->adapter, -img_len);
+      goto again;
+    }
   }
 
+  dec->rem_img_len = img_len;
+
+  GST_LOG_OBJECT (dec, "image size = %u", img_len);
+
   /* QoS: if we're too late anyway, skip decoding */
   if (dec->packetized && !gst_jpeg_dec_do_qos (dec, timestamp))
     goto skip_decoding;
 
-  GST_LOG_OBJECT (dec, "image size = %u", img_len);
-
 #ifndef GST_DISABLE_GST_DEBUG
   data = (guint8 *) gst_adapter_peek (dec->adapter, 4);
   GST_LOG_OBJECT (dec, "reading header %02x %02x %02x %02x", data[0], data[1],
       data[2], data[3]);
 #endif
 
-  dec->rem_img_len = img_len;
   gst_jpeg_dec_fill_input_buffer (&dec->cinfo);
 
   if (setjmp (dec->jerr.setjmp_buffer)) {
@@ -1397,6 +1532,11 @@ gst_jpeg_dec_chain (GstPad * pad, GstBuffer * buf)
       goto drop_buffer;
   }
 
+  /* reset error count on successful decode */
+  dec->error_count = 0;
+
+  ++dec->good_count;
+
   GST_LOG_OBJECT (dec, "pushing buffer (ts=%" GST_TIME_FORMAT ", dur=%"
       GST_TIME_FORMAT, GST_TIME_ARGS (GST_BUFFER_TIMESTAMP (outbuf)),
       GST_TIME_ARGS (GST_BUFFER_DURATION (outbuf)));
@@ -1409,6 +1549,11 @@ done:
 
 exit:
 
+  if (G_UNLIKELY (ret == GST_FLOW_ERROR)) {
+    jpeg_abort_decompress (&dec->cinfo);
+    ret = gst_jpeg_dec_post_error_or_warning (dec);
+  }
+
   return ret;
 
   /* special cases */
@@ -1425,9 +1570,8 @@ need_more_data:
   /* ERRORS */
 wrong_size:
   {
-    GST_ELEMENT_ERROR (dec, STREAM, DECODE,
-        ("Picture is too small or too big (%ux%u)", width, height),
-        ("Picture is too small or too big (%ux%u)", width, height));
+    gst_jpeg_dec_set_error (dec, GST_FUNCTION, __LINE__,
+        "Picture is too small or too big (%ux%u)", width, height);
     ret = GST_FLOW_ERROR;
     goto done;
   }
@@ -1437,8 +1581,9 @@ decode_error:
 
     dec->jerr.pub.format_message ((j_common_ptr) (&dec->cinfo), err_msg);
 
-    GST_ELEMENT_ERROR (dec, STREAM, DECODE,
-        (_("Failed to decode JPEG image")), ("Error #%u: %s", code, err_msg));
+    gst_jpeg_dec_set_error (dec, GST_FUNCTION, __LINE__,
+        "Decode error #%u: %s", code, err_msg);
+
     if (outbuf) {
       gst_buffer_unref (outbuf);
       outbuf = NULL;
@@ -1462,10 +1607,10 @@ alloc_failed:
     GST_DEBUG_OBJECT (dec, "failed to alloc buffer, reason %s", reason);
     /* Reset for next time */
     jpeg_abort_decompress (&dec->cinfo);
-    if (GST_FLOW_IS_FATAL (ret)) {
-      GST_ELEMENT_ERROR (dec, STREAM, DECODE,
-          ("Buffer allocation failed, reason: %s", reason),
-          ("Buffer allocation failed, reason: %s", reason));
+    if (ret != GST_FLOW_UNEXPECTED && ret != GST_FLOW_WRONG_STATE &&
+        ret != GST_FLOW_NOT_LINKED) {
+      gst_jpeg_dec_set_error (dec, GST_FUNCTION, __LINE__,
+          "Buffer allocation failed, reason: %s", reason);
     }
     goto exit;
   }
@@ -1478,22 +1623,22 @@ drop_buffer:
   }
 components_not_supported:
   {
-    GST_ELEMENT_ERROR (dec, STREAM, DECODE, (NULL),
-        ("more components than supported: %d > 3", dec->cinfo.num_components));
+    gst_jpeg_dec_set_error (dec, GST_FUNCTION, __LINE__,
+        "more components than supported: %d > 3", dec->cinfo.num_components);
     ret = GST_FLOW_ERROR;
     goto done;
   }
 unsupported_colorspace:
   {
-    GST_ELEMENT_ERROR (dec, STREAM, DECODE, (NULL),
-        ("Picture has unknown or unsupported colourspace"));
+    gst_jpeg_dec_set_error (dec, GST_FUNCTION, __LINE__,
+        "Picture has unknown or unsupported colourspace");
     ret = GST_FLOW_ERROR;
     goto done;
   }
 invalid_yuvrgbgrayscale:
   {
-    GST_ELEMENT_ERROR (dec, STREAM, DECODE, (NULL),
-        ("Picture is corrupt or unhandled YUV/RGB/grayscale layout"));
+    gst_jpeg_dec_set_error (dec, GST_FUNCTION, __LINE__,
+        "Picture is corrupt or unhandled YUV/RGB/grayscale layout");
     ret = GST_FLOW_ERROR;
     goto done;
   }
@@ -1506,6 +1651,10 @@ gst_jpeg_dec_src_event (GstPad * pad, GstEvent * event)
   gboolean res;
 
   dec = GST_JPEG_DEC (gst_pad_get_parent (pad));
+  if (G_UNLIKELY (dec == NULL)) {
+    gst_event_unref (event);
+    return FALSE;
+  }
 
   switch (GST_EVENT_TYPE (event)) {
     case GST_EVENT_QOS:{
@@ -1588,6 +1737,9 @@ gst_jpeg_dec_set_property (GObject * object, guint prop_id,
     case PROP_IDCT_METHOD:
       dec->idct_method = g_value_get_enum (value);
       break;
+    case PROP_MAX_ERRORS:
+      g_atomic_int_set (&dec->max_errors, g_value_get_int (value));
+      break;
 
     default:
       G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
@@ -1607,6 +1759,9 @@ gst_jpeg_dec_get_property (GObject * object, guint prop_id, GValue * value,
     case PROP_IDCT_METHOD:
       g_value_set_enum (value, dec->idct_method);
       break;
+    case PROP_MAX_ERRORS:
+      g_value_set_int (value, g_atomic_int_get (&dec->max_errors));
+      break;
 
     default:
       G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
@@ -1624,6 +1779,8 @@ gst_jpeg_dec_change_state (GstElement * element, GstStateChange transition)
 
   switch (transition) {
     case GST_STATE_CHANGE_READY_TO_PAUSED:
+      dec->error_count = 0;
+      dec->good_count = 0;
       dec->framerate_numerator = 0;
       dec->framerate_denominator = 1;
       dec->caps_framerate_numerator = dec->caps_framerate_denominator = 0;
